from __future__ import annotations

from copy import copy
from dataclasses import dataclass, field
from typing import Container, Iterable, List, Optional, Sequence, Union
import requests
import bs4
from bs4 import BeautifulSoup
from bs4.element import PageElement

WiktionaryPageStr = str  # TODO remove this and implement WiktionaryPage instead


class WiktionaryPageNotFound(LookupError):
    pass


class WiktionarySectionNotFound(LookupError):
    pass


@dataclass
class WordInfo:
    word: str
    definition_strings: List[str] = field(default_factory=list)

    @classmethod
    def from_search(cls, word: str, *, revision: Optional[int] = None):
        section = WiktionaryPage(word, revision).get_spanish_section()
        return cls(word=word, definition_strings=section.definitions())


class WiktionaryPage:
    word: Optional[str] = None
    revision: Optional[int] = None

    def __init__(self,
                 word: str,
                 revision: Optional[int] = None):
        self.word = word
        self.revision = revision
        self.soup = BeautifulSoup(
            markup=self.get_page_html(word, revision), 
            features='html.parser'
        )
    
    @staticmethod
    def get_page_html(word, revision=None):
        if revision is None:
            url = f'https://en.wiktionary.org/wiki/{word}'
        else:
            url = f'https://en.wiktionary.org/w/index.php?title={word}&oldid={revision}'
        content = request_url_text(url)
        if 'Wiktionary does not yet have an entry for' in content:
            raise WiktionaryPageNotFound('No Wiktionary page found')
        return content

    def __contains__(self, other: str) -> bool:
        return other in str(self.soup)

    def get_spanish_section(self) -> WiktionaryPageSection:
        return self.get_section(language='Spanish')
    
    def get_section(self, language: str):
        return WiktionaryPageSection(
            soup=_extract_language_section(self.soup, language=language)
        )
    

def _extract_language_section(page_soup: BeautifulSoup, language: str) -> BeautifulSoup:
    """
    Get a new BeautifulSoup object that only has the tags from the section that matches 
    `language`.
    """
    tags = _language_section_tags(page_soup, language)
    return tags_to_soup(tags)


def _language_section_tags(page_soup: BeautifulSoup, language: str) -> List[PageElement]:
    """
    Get a list of all BeautifulSoup tags in the logical section that matches `language`.
    
    TODO clarify whether these are guaranteed to be **Tag**s or if they're more general 
        **PageElement**s
    """
    start_tag = _language_section_start_tag(page_soup, language)
    return get_heading_siblings_on_level(start_tag)
    
    
def _language_section_start_tag(page_soup: BeautifulSoup, language: str) -> bs4.Tag:
    section_id_tag: bs4.Tag = page_soup.find(id=language)
    if section_id_tag is None:
        raise WiktionarySectionNotFound('No Spanish entry found from Wiktionary page')
    start_tag = section_id_tag.parent
    assert start_tag.name == 'h2'
    return start_tag


class WiktionaryPageSection:

    def __init__(self, soup: BeautifulSoup):
        self.soup = self._clean_soup(soup)

    @staticmethod
    def _clean_soup(soup: BeautifulSoup):
        """ Return a copy of soup, some elements thrown away """
        soup = copy(soup)  # TODO assert that the original soup is unchanged
        for e in soup.find_all(class_='wiktQuote'):
            e.parent.decompose()

        # remove nested lists from definition list items
        for li in WiktionaryPageSection._definition_list_items_from_soup(soup):
            for e in li.find_all('ul'):
                e.decompose()

        return soup

    def __contains__(self, other: str) -> bool:
        return other in str(self.soup)
    
    def get_subsections(self, level='h3') -> List[Subsection]:
        subheadings = self.soup.find_all(level)
        tag_sets = [get_heading_siblings_on_level(sh) for sh in subheadings]
        subsoups = [tags_to_soup(tags) for tags in tag_sets]
        return [Subsection(subsoup) for subsoup in subsoups]
    
    def get_subsection(self, title, *, level='h3') -> Subsection:
        subsections = self.get_subsections(level=level)
        for ss in subsections:
            if ss.title == title:
                return ss
        else:
            raise KeyError(f'No section with title: {title}')

    def definitions(self) -> List[str]:
        definitions_ = []
        for definition_list_item in self._definition_list_items():
            definitions_.append(
                self.definition_list_item_to_str(definition_list_item))
        return definitions_

    # TODO remove
    def _definition_list_items(self):
        return self._definition_list_items_from_soup(self.soup)

    @staticmethod
    def _definition_list_items_from_soup(soup):
        headwords = soup.find_all(class_='headword')

        # a bunch of `ol`s that contain a number of definition list items
        definition_lists = [
            hw.parent.find_next_sibling('ol') for hw in headwords
        ]

        # a bunch of `li`s that contain one definition "line"
        definition_list_items = []
        for dl in definition_lists:
            for dli in dl.find_all('li'):
                definition_list_items.append(dli)

        return definition_list_items

    @staticmethod
    def definition_list_item_to_str(li: bs4.Tag) -> str:
        """
        Parse the contents of the given definition `li` tag.
        """
        str_ = ''
        for element in li.contents:
            if element.name != 'dl':  #  exclude usage examples and synonyms (etc.?), which are under <dl>
                str_ += ''.join(element.strings)
        str_ = str_.replace('\n', '')
        return str_


class Subsection(WiktionaryPageSection):
    @property
    def title(self):
        return self.soup.find(class_='mw-headline').text
    
    def content_string(self) -> str:
        cs = ''
        for c in self.soup.contents:
            if c.name == 'p':
                cs += c.text
            if c.name == 'ol':
                cs += f'- {c.text}'
        cs += '\n'
        return cs



def get_word_info(word: str, revision: Optional[int] = None):
    return WordInfo.from_search(word=word, revision=revision)


def request_url_text(url: str) -> str:
    return requests.get(url).text  # pragma: no cover


def tags_to_soup(tags: Sequence[bs4.Tag],
                 *,
                 features='html.parser') -> BeautifulSoup:
    """
    Given a list of tags, create a new BeautifulSoup object from those tags (copying tags to the
    new soup object in order)
    """
    soup = BeautifulSoup(features=features)
    for element in tags:
        soup.append(copy(element))
    return soup


def get_heading_siblings_on_level(element):
    """
    Return a list of sibling elements until the next occurrence of a heading on the same 
    or higher level.
    """
    hierarchy = 'h1', 'h2', 'h3', 'h4', 'h5', 'h6'
    if element.name not in hierarchy:
        raise ValueError(
            f'Element with {element.name} (expected one of {hierarchy})')
    
    same_and_higher = hierarchy[:hierarchy.index(element.name) + 1]
    return get_siblings_until(element, same_and_higher)


def get_siblings_until(element: PageElement,
                       until: Union[str, Container[str]]) -> List[PageElement]:
    """
    Return a list of sibling elements until the next occurrence of a certain tag name (or 
    list/tuple/etc. of tag names).
    
    The `element` itself is included, and the found occurrence of `until` is excluded.
    """
    break_names = [until] if isinstance(until, str) else until
    found = [element]
    for sibling in element.next_siblings:
        if sibling.name in break_names:
            break
        else:
            found.append(sibling)
    return found



import json
from pathlib import Path
from textwrap import dedent

import pytest
from bs4 import BeautifulSoup
from pytest_mock import MockerFixture

import palabras.core
import palabras.cli
from palabras.core import Subsection, WiktionaryPage, WordInfo, get_heading_siblings_on_level, get_siblings_until, request_url_text


MOCK_CACHE_FILE_PATH = Path(__file__).parent / '../data/mock_cache.json'


@pytest.fixture()
def mocked_request_url_text(mocker: MockerFixture):
    """
    Mock palabras.core.request_url_text to read URL -> content (HTML) mappings from a pre-populated
    file instead of over the internet.
    """
    with open(MOCK_CACHE_FILE_PATH) as fp:
        mock_cache: dict = json.load(fp)['url_contents']
    mocker.patch('palabras.core.request_url_text', side_effect=mock_cache.get)


def test_get_word_info_return_type(mocked_request_url_text):
    word = 'despacito'
    wi = palabras.core.get_word_info(word)
    assert isinstance(wi, palabras.core.WordInfo)


def test_get_word_info_from_search_return_type(mocked_request_url_text):
    word = 'despacito'
    wi = WordInfo.from_search(word)
    assert isinstance(wi, palabras.core.WordInfo)


def test_get_word_info_equals_but_is_not_word_info_from_search(mocked_request_url_text):
    word = 'despacito'
    wi1 = palabras.core.get_word_info(word)
    wi2 = WordInfo.from_search(word)
    assert wi1 == wi2
    assert wi1 is not wi2


def test_get_wiktionary_page_nonexistent(mocked_request_url_text):
    word = 'thispageaintexistent'
    with pytest.raises(palabras.core.WiktionaryPageNotFound):
        palabras.core.WiktionaryPage(word)


def test_get_wiktionary_page_contains_translation(mocked_request_url_text):
    word = 'culpar'
    translation = 'blame'
    page = palabras.core.WiktionaryPage(word)
    assert translation in page


def test_wiktionary_page_contains_portuguese_conjugation(mocked_request_url_text):
    word = 'culpar'
    expected_contains = 'culpou'  # Portuguese 3rd person preterite
    page = palabras.core.WiktionaryPage(word)
    assert expected_contains in page


def test_spanish_section_type(mocked_request_url_text):
    word = 'culpar'
    result = WiktionaryPage(word).get_spanish_section()
    assert isinstance(result, palabras.core.WiktionaryPageSection)


def test_no_spanish_definition(mocked_request_url_text):
    word = 'kauppa'  # a word that has Wiktionary page but no Spanish definition
    with pytest.raises(palabras.core.WiktionarySectionNotFound):
        WiktionaryPage(word).get_spanish_section()


def test_spanish_section_does_not_contain_portuguese(mocked_request_url_text):
    word = 'culpar'
    portuguese_conjugation = 'culpou'  # Portuguese 3rd person preterite
    section = WiktionaryPage(word).get_spanish_section()
    assert portuguese_conjugation not in section


def test_definition_list_item_to_str(mocked_request_url_text):
    li = BeautifulSoup('''
    <li>parse <a href="foo">this</a><dl><dd><span>Whatever</span>...</dd></dl></li>
    ''', features='html.parser').li
    str_definition = palabras.core.WiktionaryPageSection.definition_list_item_to_str(li)
    assert str_definition == 'parse this'


def test_lookup_definition(mocked_request_url_text):
    word = 'culpar'
    word_info = palabras.core.get_word_info(word)
    assert word_info.definition_strings[0] == 'to blame'


def test_lookup_definition_complicated(mocked_request_url_text):
    word = 'empleado'  # this word has definitions for adjective, noun, and verb
    revision = 62175311
    word_info = palabras.core.get_word_info(word, revision=revision)
    assert word_info.definition_strings == [
        'employed',
        'employee',
        'Masculine singular past participle of emplear.'
    ]


def test_lookup_different_definitions_in_history(mocked_request_url_text):
    word = 'olvidar'
    revision_1 = 62345284
    revision_2 = 66217360
    expected_definitions_1 = [
        'to forget; to elude, escape (be forgotten by)',
        '(reflexive) to forget',
        '(reflexive) to leave behind'
    ]
    expected_definitions_2 = [
        '(transitive) to forget (be forgotten by)',
        '(reflexive, intransitive) to forget, elude, escape',
        '(with de, reflexive, intransitive) to forget, to leave behind'
    ]

    assert palabras.core.get_word_info(word, revision=revision_1).definition_strings \
        == expected_definitions_1
    assert palabras.core.get_word_info(word, revision=revision_2).definition_strings \
        == expected_definitions_2

    
def test_cli(capsys: pytest.CaptureFixture, mocked_request_url_text):
    args = ['olvidar']
    palabras.cli.main(args)
    captured = capsys.readouterr()
    expected = dedent('''
        olvidar
        - (transitive) to forget (be forgotten by)
        - (reflexive, intransitive) to forget, elude, escape
        - (with de, reflexive, intransitive) to forget, to leave behind
    ''').lstrip()
    assert captured.out == expected
    

def test_cli_revision(capsys: pytest.CaptureFixture, mocked_request_url_text):
    args1 = ['olvidar', '-r', '62345284']
    args2 = ['olvidar', '--revision', '62345284']
    args3 = ['olvidar', '--revision=62345284']
    
    expected = dedent('''
        olvidar
        - to forget; to elude, escape (be forgotten by)
        - (reflexive) to forget
        - (reflexive) to leave behind
    ''').lstrip()
    
    for args in args1, args2, args3:
        exitcode = palabras.cli.main(args)
        captured = capsys.readouterr()
        assert captured.out == expected
        assert exitcode == 0
    

def test_cli_ser(capsys: pytest.CaptureFixture, mocked_request_url_text):
    args = ['ser']
    exitcode = palabras.cli.main(args)
    captured = capsys.readouterr()
    expected = dedent('''
        ser
        - to be (essentially or identified as)
        - to be (in the passive voice sense)
        - to exist; to occur
        - a being, organism
        - nature, essence
        - value, worth
    ''').lstrip()
    assert captured.out == expected
    assert exitcode == 0
    

def test_cli_nonexistent_page(capsys: pytest.CaptureFixture, mocked_request_url_text):
    args = ['asdasdasd']
    exitcode = palabras.cli.main(args)
    captured = capsys.readouterr()
    expected = 'No Wiktionary page found\n'
    assert captured.out == expected
    assert exitcode == 1
    

def test_cli_non_spanish_section(capsys: pytest.CaptureFixture, mocked_request_url_text):
    args = ['moikka']
    exitcode = palabras.cli.main(args)
    captured = capsys.readouterr()
    expected = 'No Spanish entry found from Wiktionary page\n'
    assert captured.out == expected
    assert exitcode == 1


def test_get_next_siblings_until():
    markup = (
        '<h1>hello</h1>'
        '<p>foo</p>'
        '<h2>hello again</h2>'
        '<p>bar</p>'
        '<h1>hello 2</h1>'
        '<p>foo 2</p>'
        '<h2>hello again 2</h2>'
        '<p>bar 2</p>'
        '<h1>hello 3</h1>'
    )
    soup = BeautifulSoup(markup, features='html.parser')
    tag = soup.find('h1')
    assert len(get_siblings_until(tag, 'p')) == 1  # hello
    assert len(get_siblings_until(tag, 'h2')) == 2  # hello ... hello again
    assert len(get_siblings_until(tag, 'h1')) == 4  # hello ... hello 2
    assert len(get_siblings_until(tag, 'nope')) == 9  # all tags
    
    tag2 = soup.find_all('h1')[1]
    assert len(get_siblings_until(tag2, 'h1')) == 4  # hello 2 ... bar 2
    assert len(get_siblings_until(tag2, 'nope')) == 5  # hello 2 ... hello 3
    
    
def test_page_object():
    word = 'ser'
    page = WiktionaryPage(word)
    assert isinstance(page, WiktionaryPage)
    
    
def test_page_object_from_word_and_revision():
    word = 'empleado'
    revision = 62175311
    page = WiktionaryPage(word, revision)
    assert isinstance(page, WiktionaryPage)
    
    
def test_page_object_attributes():
    word = 'empleado'
    revision = 62175311
    page = WiktionaryPage(word, revision)
    assert page.word == word
    assert page.revision == revision


def test_get_subsections_len_and_type():
    page = WiktionaryPage('empleado')
    section = page.get_spanish_section()
    subsections = section.get_subsections()
    assert len(subsections) > 0  # this page has sections
    for subsection in subsections:
        assert isinstance(subsection, Subsection)


def test_get_subsections_titles():
    page = WiktionaryPage('empleado', revision=68396093)
    section = page.get_spanish_section()
    subsections = section.get_subsections()
    titles = [ss.title for ss in subsections]
    assert titles == [
        'Etymology',
        'Pronunciation',
        'Adjective',
        'Noun',
        'Participle',
        'Further reading'
    ]
    
    
def test_get_specific_subsection():
    page = WiktionaryPage('empleado')
    section = page.get_spanish_section()
    subsection_adjective = section.get_subsection('Adjective')
    assert isinstance(subsection_adjective, Subsection)
    assert subsection_adjective.title == 'Adjective'
    
    
def test_get_nonexistent_subsection():
    page = WiktionaryPage('empleado')
    section = page.get_spanish_section()
    with pytest.raises(KeyError, match='No section with title:'):
        section.get_subsection('Nonexistent section')
        
        
def test_adjective_subsection_content_string():
    page = WiktionaryPage('empleado')
    expected = dedent('''
        empleado (feminine empleada, masculine plural empleados, feminine plural empleadas)
        - employed
    ''').lstrip()
    ss = page.get_spanish_section().get_subsection('Adjective')
    assert ss.content_string() == expected
        
        
def test_get_heading_siblings_on_level():
    soup = BeautifulSoup(
        '<h1>tag 1</h1>'
        '<h2>tag 2</h2>'
        '<h3>tag 3</h3>'
        '<p>tag 4</p>'
        '<h3>tag 5</h3>'
        '<h2>tag 6</h2>',
        features='html.parser'
    )
    element = soup.h2  # find the first h2
    # should find 2,3,4,5
    assert len(get_heading_siblings_on_level(element)) == 4
    
    element = soup.h3 # find the first h3
    # should find 3,4
    assert len(get_heading_siblings_on_level(element)) == 2


def test_get_siblings_on_level_error_on_unexpected_element():
    soup = BeautifulSoup(
        '<h1>one</h1>'
        '<li>get_siblings_on_level</li>'
        '<li>only works for headings</li>',
        features='html.parser'
    )
    element = soup.li
    with pytest.raises(ValueError):
        get_heading_siblings_on_level(element)


import argparse

from .core import WordInfo, WiktionaryPageNotFound, WiktionarySectionNotFound


def main(args):
    formatter = lambda prog: argparse.HelpFormatter(prog, max_help_position=50)
    parser = argparse.ArgumentParser(description='Look up a word', formatter_class=formatter)
    parser.add_argument('word', metavar='<word>', type=str, help='A word to look up')
    parser.add_argument('-r', '--revision', type=int, metavar='<n>', help='Wiktionary revision ID (from permalink)')
    args = parser.parse_args(args)
    
    try:
        print(parse(WordInfo.from_search(args.word, revision=args.revision)))
        return 0
    except (WiktionaryPageNotFound, WiktionarySectionNotFound) as exc:
        print(exc)
        return 1


def parse(word_info: WordInfo) -> str:
    lines = [word_info.word]
    lines += [f'- {ds}' for ds in word_info.definition_strings]
    return '\n'.join(lines)